"""This package comes with a built-in CLI for running a resolver web application.

.. code-block::

    $ python -m curies resolver --host 0.0.0.0 --port 8764 bioregistry

The positional argument can be one of the following:

1. A pre-defined prefix map to get from the web (bioregistry, go, obo, monarch, prefixcommons)
2. A local file path or URL to a prefix map, extended prefix map, or one of several formats. Requires specifying
   a `--format`.

The framework can be swapped to use Flask (default) or FastAPI with `--framework`. The
server can be swapped to use Werkzeug (default) or Uvicorn with `--server`. These functionalities
are also available programmatically (see :func:`get_flask_app` and :func:`get_fastapi_app`).

Similarly, there's a built-in CLI for running a mapper web application.

.. code-block::

    $ python -m curies mapper --host 0.0.0.0 --port 8764 bioregistry

The same flags and arguments are applicable.
"""

from __future__ import annotations

import sys
from collections.abc import Mapping
from typing import TYPE_CHECKING, Callable

import click
from typing_extensions import TypeAlias

from curies import Converter, sources

if TYPE_CHECKING:
    import fastapi
    import flask

__all__ = [
    "main",
]

AppHint: TypeAlias = "flask.Flask | fastapi.FastAPI"


LOADERS: dict[str, Callable[[str], Converter]] = {
    "jsonld": Converter.from_jsonld,
    "prefix_map": Converter.from_prefix_map,
    "extended_prefix_map": Converter.from_extended_prefix_map,
    "reverse_prefix_map": Converter.from_reverse_prefix_map,
    "priority_prefix_map": Converter.from_priority_prefix_map,
}

CONVERTERS: Mapping[str, Callable[[], Converter]] = {
    "bioregistry": sources.get_bioregistry_converter,
    "go": sources.get_go_converter,
    "monarch": sources.get_monarch_converter,
    "obo": sources.get_obo_converter,
    "prefixcommons": sources.get_prefixcommons_converter,
}


def _get_converter(location: str, format: str | None) -> Converter:
    if location in CONVERTERS:
        return CONVERTERS[location]()
    if format is None:
        click.secho("--format is required with remote data", fg="red")
        return sys.exit(1)
    return LOADERS[format](location)


def _get_resolver_app(converter: Converter, framework: str) -> AppHint:
    from curies import resolver_service

    if framework == "flask":
        return resolver_service.get_flask_app(converter)
    elif framework == "fastapi":
        return resolver_service.get_fastapi_app(converter)
    else:
        raise ValueError(f"Unhandled framework: {framework}")


def _get_mapper_app(converter: Converter, framework: str) -> AppHint:
    from curies import mapping_service

    if framework == "flask":
        return mapping_service.get_flask_mapping_app(converter)
    elif framework == "fastapi":
        return mapping_service.get_fastapi_mapping_app(converter)
    else:
        raise ValueError(f"Unhandled framework: {framework}")


def _run_app(app: AppHint, server: str, host: str, port: int) -> None:
    if server == "uvicorn":
        import uvicorn

        uvicorn.run(app, host=host, port=port)
    elif server == "werkzeug":
        # we ignore the type because at this point, we know the app has to be a flask.Flask
        app.run(host=host, port=port)  # type:ignore[union-attr]
    elif server == "gunicorn":
        raise NotImplementedError
    else:
        raise ValueError(f"Unhandled server: {server}")


LOCATION_ARGUMENT = click.argument("location")
FRAMEWORK_OPTION = click.option(
    "--framework",
    default="flask",
    type=click.Choice(["flask", "fastapi"]),
    show_default=True,
    help="The framework used to implement the app. Note, each requires different packages to be installed.",
)
SERVER_OPTION = click.option(
    "--server",
    default="werkzeug",
    type=click.Choice(["werkzeug", "uvicorn", "gunicorn"]),
    show_default=True,
    help="The web server used to run the app. Note, each requires different packages to be installed.",
)
FORMAT_OPTION = click.option(
    "--format",
    type=click.Choice(list(LOADERS)),
    help="The data structure of the resolver data. Required if not giving a pre-defined converter name.",
)
HOST_OPTION = click.option(
    "--host",
    default="0.0.0.0",  # noqa:S104
    show_default=True,
    help="The host where the resolver runs",
)
PORT_OPTION = click.option(
    "--port", type=int, default=8764, show_default=True, help="The port where the resolver runs"
)


@click.group()
def main() -> None:
    """Run the `curies` CLI."""


@main.command(
    help="""\
Location can either be the name of a built-in converter, a file path, or a URL.
"""
)
@LOCATION_ARGUMENT
@FRAMEWORK_OPTION
@SERVER_OPTION
@FORMAT_OPTION
@HOST_OPTION
@PORT_OPTION
def resolver(location: str, host: str, port: int, framework: str, format: str, server: str) -> None:
    """Serve a resolver app."""
    converter = _get_converter(location, format)
    app = _get_resolver_app(converter, framework=framework)
    _run_app(app, server=server, host=host, port=port)


@main.command(
    help="""\
Location can either be the name of a built-in converter, a file path, or a URL.
"""
)
@LOCATION_ARGUMENT
@FRAMEWORK_OPTION
@SERVER_OPTION
@FORMAT_OPTION
@HOST_OPTION
@PORT_OPTION
def mapper(location: str, host: str, port: int, framework: str, format: str, server: str) -> None:
    """Serve a mapper app."""
    converter = _get_converter(location, format)
    app = _get_mapper_app(converter, framework=framework)
    _run_app(app, server=server, host=host, port=port)


if __name__ == "__main__":
    main()



"""Tests for the simple web service."""

import unittest
from typing import ClassVar

from fastapi.testclient import TestClient

from curies import Converter
from curies.resolver_service import FAILURE_CODE, get_fastapi_app, get_flask_app


class ConverterMixin(unittest.TestCase):
    """A mixin that has a converter."""

    delimiter: ClassVar[str] = ":"

    def setUp(self) -> None:
        """Set up the test case with a converter."""
        super().setUp()
        self.converter = Converter.from_prefix_map(
            {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                "MONDO": "http://purl.obolibrary.org/obo/MONDO_",
                "GO": "http://purl.obolibrary.org/obo/GO_",
                "OBO": "http://purl.obolibrary.org/obo/",
            },
            delimiter=self.delimiter,
        )


class TestFastAPI(ConverterMixin):
    """Test building a simple web service with FastAPI."""

    def setUp(self) -> None:
        """Set up the test case with a converter, blueprint, and app."""
        super().setUp()
        self.app = get_fastapi_app(self.converter)
        self.client = TestClient(self.app)

    def test_resolve_success(self) -> None:
        """Test resolution for a valid CURIE redirects properly."""
        curie = self.converter.format_curie("GO", "1234567")
        res = self.client.get(f"/{curie}", follow_redirects=False)
        self.assertEqual(302, res.status_code, msg=res.text)

    def test_resolve_failure(self) -> None:
        """Test resolution for an invalid CURIE aborts with 404."""
        curie = self.converter.format_curie("NOPREFIX", "NOIDENTIFIER")
        res = self.client.get(f"/{curie}", follow_redirects=False)
        self.assertEqual(FAILURE_CODE, res.status_code, msg=res.text)


class TestFastAPISlashed(TestFastAPI):
    """Test the FastAPI router with an alternate delimiter."""

    delimiter = "/"

    def test_delimiter(self) -> None:
        """Test the delimiter."""
        self.assertEqual("/", self.converter.delimiter)


class TestFlaskBlueprint(ConverterMixin):
    """Test building a simple web service with Flask."""

    def setUp(self) -> None:
        """Set up the test case with a converter, blueprint, and app."""
        super().setUp()
        self.app = get_flask_app(self.converter)

    def test_resolve_success(self) -> None:
        """Test resolution for a valid CURIE redirects properly."""
        curie = self.converter.format_curie("GO", "1234567")
        with self.app.test_client() as client:
            res = client.get(f"/{curie}", follow_redirects=False)
            self.assertEqual(302, res.status_code, msg=res.text)

    def test_resolve_failure(self) -> None:
        """Test resolution for an invalid CURIE aborts with 404."""
        curie = self.converter.format_curie("NOPREFIX", "NOIDENTIFIER")
        with self.app.test_client() as client:
            res = client.get(f"/{curie}", follow_redirects=False)
            self.assertEqual(FAILURE_CODE, res.status_code, msg=res.text)


class TestFlaskBlueprintSlashed(TestFlaskBlueprint):
    """Test the flask blueprint with an alternate delimiter."""

    delimiter = "/"

    def test_delimiter(self) -> None:
        """Test the delimiter."""
        self.assertEqual("/", self.converter.delimiter)



"""Discovery new entries for a Converter.

The :func:`curies.discover` functionality is intended to be used in a "data science" workflow. Its goal is to
enable a data scientist to semi-interactively explore data (e.g., coming from an ontology, SSSOM, RDF)
that doesn't come with a complete (extended) prefix map and identify common URI prefixes.

It returns the discovered URI prefixes in a :class:`curies.Converter` object with "dummy" CURIE prefixes.
This makes it possible to convert the URIs appearing in the data into CURIEs and therefore enables
their usage in places where CURIEs are expected.

However, it's suggested that after discovering URI prefixes, the data scientist more carefully
constructs a meaningful prefix map based on the discovered one. This might include some or all of the
following steps:

1. Replace dummy CURIE prefixes with meaningful ones
2. Remove spurious URI prefixes that appear but do not represent a semantic space. This happens often
   due to using ``_`` as a delimiter or having a frequency cutoff of zero (see the parameters for this function).
3. Consider chaining a comprehensive extended prefix map such as the Bioregistry
   (from :func:`curies.get_bioregistry_converter`) with onto the converter passed to this function
   so pre-existing URI prefixes are not *re-discovered*.

Finally, you should save the prefix map that you create in a persistent place (i.e., inside a JSON file) such
that it can be reused.

Algorithm
---------
The :func:`curies.discover` function implements the following algorithm that does the following for each URI:

1. For each delimiter (in the priority order they are given) check if the delimiter is present.
2. If it's present, split the URI into two parts based on rightmost appearance of the delimiter.
3. If the right part after splitting is all alphanumeric characters, save the URI prefix (with delimiter attached)
4. If a delimiter is successfully used to identify a URI prefix, don't check any of the following delimiters

After identifying putative URI prefixes, the second part of the algorithm does the following:

1. If a cutoff was provided, remove all putative URI prefixes for which there were fewer examples than the cutoff
2. Sort the URI prefixes lexicographically (i.e., with :func:`sorted`)
3. Assign a dummy CURIE prefix to each URI prefix, counting upwards from 1
4. Construct a converter from this prefix map and return it
"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from pathlib import PurePath
from typing import IO, TYPE_CHECKING, Any, TextIO, Union

from typing_extensions import Literal

from curies import Converter, Record

if TYPE_CHECKING:
    import rdflib

__all__ = [
    "discover",
    "discover_from_rdf",
]


GraphFormats = Literal["turtle", "xml", "n3", "nt", "trix"]
GraphInput = Union[IO[bytes], TextIO, "rdflib.parser.InputSource", str, bytes, PurePath]


def discover_from_rdf(
    graph: GraphInput | rdflib.Graph,
    *,
    format: GraphFormats | None = None,
    **kwargs: Any,
) -> Converter:
    """Discover new URI prefixes from RDF content via :mod:`rdflib`.

    This function works the same as :func:`discover`, but gets its URI list from
    a triple store. See :func:`discover` for a more detailed explanation of how this
    algorithm works.

    :param graph:
        Either a pre-instantiated RDFlib graph, or an input type to the ``source``
        keyword of :meth:`rdflib.Graph.parse`. This can be one of the following:

        - A string or bytes representation of a URL
        - A string, bytes, or Path representation of a local file
        - An I/O object that can be read directly
        - An open XML reader from RDFlib (:class:`rdflib.parser.InputSource`)
    :param format: If ``graph`` is given as a URL or I/O object, this
        is passed through to the ``format`` keyword of :meth:`rdflib.Graph.parse`.
        If none is given, defaults to ``turtle``.
    :param kwargs: Keyword arguments passed through to :func:`discover`
    :returns:
        A converter with dummy prefixes for URI prefixes appearing in the RDF
        content (i.e., triples).
    """
    uris = get_uris_from_rdf(graph=graph, format=format)
    return discover(uris, **kwargs)


def get_uris_from_rdf(
    graph: GraphInput | rdflib.Graph, *, format: GraphFormats | None = None
) -> set[str]:
    """Get a set of URIs from a graph."""
    graph = _ensure_graph(graph=graph, format=format)
    return set(_yield_uris(graph=graph))


def _ensure_graph(
    *, graph: GraphInput | rdflib.Graph, format: GraphFormats | None = None
) -> rdflib.Graph:
    import rdflib

    if not isinstance(graph, rdflib.Graph):
        _temp_graph = rdflib.Graph()
        _temp_graph.parse(source=graph, format=format)
        graph = _temp_graph

    return graph


def _yield_uris(*, graph: rdflib.Graph) -> Iterable[str]:
    import rdflib

    for parts in graph.triples((None, None, None)):
        for part in parts:
            if isinstance(part, rdflib.URIRef):
                yield str(part)


def discover(
    uris: Iterable[str],
    *,
    delimiters: Sequence[str] | None = None,
    cutoff: int | None = None,
    metaprefix: str = "ns",
    converter: Converter | None = None,
) -> Converter:
    """Discover new URI prefixes and construct a converter with a unique dummy CURIE prefix for each.

    :param uris:
        An iterable of URIs to search through. Will be taken as a set and
        each unique entry is only considered once.
    :param delimiters:
        The character(s) that delimit a URI prefix from a local unique identifier.
        If none given, defaults to using ``/``, ``#``, and ``_``. For example:

        - ``/`` is the delimiter in ``https://www.ncbi.nlm.nih.gov/pubmed/37929212``, which separates
          the URI prefix ``https://www.ncbi.nlm.nih.gov/pubmed/`` from the local unique identifier
          `37929212 <https://www.ncbi.nlm.nih.gov/pubmed/37929212>`_ for the article
          "New insights into osmobiosis and chemobiosis in tardigrades" in PubMed.
        - ``#`` is the delimiter in ``http://www.w3.org/2000/01/rdf-schema#label``, which separates
          the URI prefix ``http://www.w3.org/2000/01/rdf-schema#`` from the local unique identifier
          `label <http://www.w3.org/2000/01/rdf-schema#label>`_ for the term "label" in the RDF Schema.
          The ``#`` typically is used in a URL to denote a fragment and commonly appears in small semantic
          web vocabularies that are shown as a single HTML page.
        - ``_`` is the delimiter in ``http://purl.obolibrary.org/obo/GO_0032571``, which separates
          the URI prefix ``http://purl.obolibrary.org/obo/GO_`` from the local unique identifier
          `0032571 <http://purl.obolibrary.org/obo/GO_0032571>`_ for the term "response to vitamin K"
          in the Gene Ontology

        .. note:: The delimiter is itself a part of the URI prefix
    :param cutoff:
        If given, will require more than ``cutoff`` unique local unique identifiers
        associated with a given URI prefix to keep it.

        Defaults to zero, which increases recall (i.e., likelihood of getting all
        possible URI prefixes) but decreases precision (i.e., more of the results
        might be false positives / spurious). If you get a lot of false positives,
        try increasing first to 1, 2, then maybe higher.
    :param metaprefix:
        The beginning part of each dummy prefix, followed by a number. The default value
        is ``ns``, so dummy prefixes are named ``ns1``, ``ns2``, and so on.
    :param converter:
        If a pre-existing converter is passed, then URIs that can be parsed using the
        pre-existing converter are not considered during discovery.

        For example, if you're an OBO person working with URIs coming from an OBO ontology,
        it makes sense to pass the converter from :func:`curies.get_obo_converter` to
        reduce false positive discoveries. More generally, a comprehensive converter
        like the Bioregistry (from :func:`curies.get_bioregistry_converter`) can massively
        reduce false positive discoveries and ultimately reduce burden on the data scientist
        using this function when needing to understand the results and carefully curate a prefix
        map based on the discoveries.
    :returns:
        A converter with dummy prefixes

    .. code-block:: python

        >>> import curies

        # Generate some example URIs
        >>> uris = [f"http://ran.dom/{i:03}" for i in range(30)]

        >>> discovered_converter = curies.discover(uris)
        >>> discovered_converter.records
        [Record(prefix="ns1", uri_prefix="http://ran.dom/")]

        # Now, you can compress the URIs to dummy CURIEs
        >>> discovered_converter.compress("http://ran.dom/002")
        'ns1:002'
    """
    uri_prefix_to_luids = _get_uri_prefix_to_luids(
        converter=converter, uris=uris, delimiters=delimiters
    )
    uri_prefixes = [
        uri_prefix
        for uri_prefix, luids in sorted(uri_prefix_to_luids.items())
        # If the cutoff is 5, and only 3 unique LUIDs with the URI prefix
        # were identified, we're going to disregard this URI prefix.
        if cutoff is None or len(luids) >= cutoff
    ]
    records = [
        Record(prefix=f"{metaprefix}{uri_prefix_index}", uri_prefix=uri_prefix)
        for uri_prefix_index, uri_prefix in enumerate(uri_prefixes, start=1)
    ]
    return Converter(records)


#: The default delimiters used when guessing URI prefixes
DEFAULT_DELIMITERS = ("#", "/", "_")


def _get_uri_prefix_to_luids(
    *,
    converter: Converter | None = None,
    uris: Iterable[str],
    delimiters: Sequence[str] | None = None,
) -> Mapping[str, set[str]]:
    """Get a mapping from putative URI prefixes to corresponding putative local unique identifiers.

    :param converter:
        A converter with pre-existing definitions. URI prefixes
        are considered "new" if they can't already be validated by this converter
    :param uris:
        An iterable of URIs to search through. Will be taken as a set and
        each unique entry is only considered once.
    :param delimiters:
        The delimiters considered between a putative URI prefix and putative
        local unique identifier. By default, checks ``#`` first since this is
        commonly used for URL fragments, then ``/`` since many URIs are constructed
        with these.
    :returns:
        A dictionary of putative URI prefixes to sets of putative local unique identifiers
    """
    if not delimiters:
        delimiters = DEFAULT_DELIMITERS
    uri_prefix_to_luids = defaultdict(set)
    for uri in uris:
        if converter is not None and converter.is_uri(uri):
            continue
        if uri.startswith("https://github.com") and "issues" in uri:
            # TODO it's not really the job of :mod:`curies` to incorporate special cases,
            #  but the GitHub thing is such an annoyance...
            continue
        for delimiter in delimiters:
            if delimiter not in uri:
                continue
            uri_prefix, luid = uri.rsplit(delimiter, maxsplit=1)
            if luid.isalnum():
                uri_prefix_to_luids[uri_prefix + delimiter].add(luid)
                break
    return dict(uri_prefix_to_luids)



# type: ignore

"""A custom SPARQL processor that optimizes the query based on https://github.com/RDFLib/rdflib/pull/2257."""

from __future__ import annotations

from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.evaluate import evalQuery
from rdflib.plugins.sparql.parser import parseQuery
from rdflib.plugins.sparql.parserutils import CompValue
from rdflib.plugins.sparql.processor import SPARQLProcessor
from rdflib.plugins.sparql.sparql import Query

__all__ = ["MappingServiceSPARQLProcessor"]


class MappingServiceSPARQLProcessor(SPARQLProcessor):
    """A custom SPARQL processor that optimizes the query based on https://github.com/RDFLib/rdflib/pull/2257.

    Why is this necessary? Ideally, we get queries like

    .. code-block:: sparql

        SELECT * WHERE {
            VALUES ?s { :a :b ... }
            ?s owl:sameAs ?o
        }

    This is fine, since the way that RDFLib parses and constructs an abstract syntax tree, the values
    for ``?s`` get bound properly when calling a custom :func:`rdflib.Graph.triples`. However, it's also
    valid SPARQL to have the ``VALUES`` clause outside of the ``WHERE`` clause like

    .. code-block:: sparql

        SELECT * WHERE {
            ?s owl:sameAs ?o
        }
        VALUES ?s { :a :b ... }

    Unfortunately, this trips up RDFLib since it doesn't know to bind the values before calling ``triples()``,
    therefore thwarting our custom implementation that dynamically generates triples based on the bound values
    themselves.

    This processor, originally by Jerven Bolleman in https://github.com/RDFLib/rdflib/pull/2257,
    adds some additional logic between parsing + constructing the abstract syntax tree and evaluation
    of the syntax tree. Basically, the abstract syntax tree has nodes with two or more children. Jerven's
    clever code (see :func:`_optimize_node` below) finds *Join* nodes that have a ``VALUES`` clause in the
    second of its two arguments, then flips them around. It does this recursively for the whole tree.
    This gets us to the goal of having the ``VALUES`` clauses appear first, therefore making sure that their
    bound values are available to the ``triples`` function.
    """

    def query(
        self,
        query: str | Query,
        initBindings=None,  # noqa:N803
        initNs=None,  # noqa:N803
        base=None,
        DEBUG=False,  # noqa:N803
    ):
        """Evaluate a SPARQL query on this processor's graph."""
        if isinstance(query, str):
            parse_tree = parseQuery(query)
            query = translateQuery(parse_tree, base, initNs)
            return self.query(query, initBindings=initBindings, base=base)

        query.algebra = _optimize_node(query.algebra)
        return evalQuery(self.graph, query, initBindings or {}, base)


# From Jerven's PR to RDFLib (https://github.com/RDFLib/rdflib/pull/2257)
def _optimize_node(comp_value: CompValue) -> CompValue:
    if (
        comp_value.name == "Join"
        and comp_value.p1.name != "ToMultiSet"
        and comp_value.p2.name == "ToMultiSet"
    ):
        comp_value.update(p1=comp_value.p2, p2=comp_value.p1)
    for inner_comp_value in comp_value.values():
        if isinstance(inner_comp_value, CompValue):
            _optimize_node(inner_comp_value)
    return comp_value



"""A module containing pre-defined references."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

from typing_extensions import TypeAlias

from curies import NamedReference as Reference

# RDF Schema

has_label = Reference(prefix="rdfs", identifier="label", name="has label")
see_also = Reference(prefix="rdfs", identifier="seeAlso", name="see also")
has_comment = Reference(prefix="rdfs", identifier="comment", name="comment")

# Synonyms

## Synonym Relations

#: The parent property for having synonyms, see :data:`synonym_scopes`
#: for a dictionary of all synonym types
has_synonym = Reference(prefix="oboInOwl", identifier="hasSynonym", name="has synonym")

has_exact_synonym = Reference(
    prefix="oboInOwl", identifier="hasExactSynonym", name="has exact synonym"
)
has_narrow_synonym = Reference(
    prefix="oboInOwl", identifier="hasNarrowSynonym", name="has narrow synonym"
)
has_broad_synonym = Reference(
    prefix="oboInOwl", identifier="hasBroadSynonym", name="has broad synonym"
)
has_related_synonym = Reference(
    prefix="oboInOwl", identifier="hasRelatedSynonym", name="has related synonym"
)

#: A list of strings used to refer to synonym types in ``oboInOwl``
SynonymScope: TypeAlias = Literal["EXACT", "NARROW", "BROAD", "RELATED"]

#: A mapping from synonym scopes to references
synonym_scopes: dict[SynonymScope, Reference] = {
    "EXACT": has_exact_synonym,
    "NARROW": has_narrow_synonym,
    "BROAD": has_broad_synonym,
    "RELATED": has_related_synonym,
}

## OMO Synonym Types

#: The parent class for all synonym types, see :data:`synonym_types`
#: for the set of all synonym types
synonym_type = Reference(prefix="oboInOwl", identifier="SynonymType", name="synonym type")

abbreviation = Reference(prefix="OMO", identifier="0003000", name="abbreviation")
ambiguous_synonym = Reference(prefix="OMO", identifier="0003001", name="ambiguous synonym")
dubious_synonym = Reference(prefix="OMO", identifier="0003002", name="dubious synonym")
layperson_synonym = Reference(prefix="OMO", identifier="0003003", name="layperson synonym")
plural_form = Reference(prefix="OMO", identifier="0003004", name="plural form")
uk_spelling = Reference(prefix="OMO", identifier="0003005", name="UK spelling synonym")
misspelling = Reference(prefix="OMO", identifier="0003006", name="misspelling")
misnomer = Reference(prefix="OMO", identifier="0003007", name="misnomer")
previous_name = Reference(prefix="OMO", identifier="0003008", name="previous name")
legal_name = Reference(prefix="OMO", identifier="0003009", name="legal name")
inn = Reference(prefix="OMO", identifier="0003010", name="International Nonproprietary Name")
latin_term = Reference(prefix="OMO", identifier="0003011", name="latin term")
acronym = Reference(prefix="OMO", identifier="0003012", name="acronym")
#: Provisional, see https://github.com/information-artifact-ontology/ontology-metadata/pull/162/files
brand_name = Reference(prefix="OMO", identifier="0003013", name="brand name")

#: A set of synonym types from OMO
synonym_types: set[Reference] = {
    abbreviation,
    ambiguous_synonym,
    dubious_synonym,
    layperson_synonym,
    plural_form,
    uk_spelling,
    misspelling,
    misnomer,
    previous_name,
    legal_name,
    inn,
    latin_term,
    acronym,
    brand_name,
}

# Semantic Mappings

## Mapping Relations

#: The parent property for semantic mappings from ``skos``, see :data:`semantic_mapping_scopes`
#: for a list of all usable properties
match = Reference(prefix="skos", identifier="mappingRelation", name="is in mapping relation with")

has_dbxref = Reference(
    prefix="oboInOwl", identifier="hasDbXref", name="has database cross-reference"
)
same_as = Reference(prefix="owl", identifier="sameAs", name="same as")
equivalent_class = Reference(prefix="owl", identifier="equivalentClass", name="equivalent class")
equivalent_property = Reference(
    prefix="owl", identifier="equivalentProperty", name="equivalent property"
)

exact_match = Reference(prefix="skos", identifier="exactMatch", name="exact match")
narrow_match = Reference(prefix="skos", identifier="narrowMatch", name="narrow match")
broad_match = Reference(prefix="skos", identifier="broadMatch", name="broad match")
close_match = Reference(prefix="skos", identifier="closeMatch", name="close match")
related_match = Reference(prefix="skos", identifier="relatedMatch", name="related match")

# `alternate term` and `term replaced by` can also be considered as mapping relations
alternative_term = Reference(prefix="IAO", identifier="0000118", name="alternative term")
term_replaced_by = Reference(prefix="IAO", identifier="0100001", name="term replaced by")

#: A list of strings used to refer to mapping types in ``skos``
SemanticMappingScope: TypeAlias = Literal["EXACT", "NARROW", "BROAD", "CLOSE", "RELATED"]

#: A mapping from mapping types to references
semantic_mapping_scopes: dict[SemanticMappingScope, Reference] = {
    "EXACT": exact_match,
    "NARROW": narrow_match,
    "BROAD": broad_match,
    "CLOSE": close_match,
    "RELATED": related_match,
}

## Matching Process Types

#: The parent class for matching processes, see :data:`matching_processes`
#: for the set of all matching processes.
matching_process = Reference(prefix="semapv", identifier="Matching", name="matching process")

background_knowledge_based_matching_process = Reference(
    prefix="semapv",
    identifier="BackgroundKnowledgeBasedMatching",
    name="background knowledge-based matching process",
)
composite_matching_process = Reference(
    prefix="semapv", identifier="CompositeMatching", name="composite matching process"
)
instance_based_matching_process = Reference(
    prefix="semapv", identifier="InstanceBasedMatching", name="instance-based matching process"
)
lexical_matching_process = Reference(
    prefix="semapv", identifier="LexicalMatching", name="lexical matching process"
)
lexical_similarity_threshold_based_matching_process = Reference(
    prefix="semapv",
    identifier="LexicalSimilarityThresholdMatching",
    name="lexical similarity threshold-based matching process",
)
logical_reasoning_matching_process = Reference(
    prefix="semapv", identifier="LogicalReasoning", name="logical reasoning matching process"
)
manual_mapping_curation = Reference(
    prefix="semapv", identifier="ManualMappingCuration", name="manual mapping curation"
)
mapping_chaining = Reference(
    prefix="semapv", identifier="MappingChaining", name="mapping chaining-based matching process"
)
mapping_inversion = Reference(
    prefix="semapv", identifier="MappingInversion", name="mapping inversion-based matching process"
)
semantic_similarity = Reference(
    prefix="semapv",
    identifier="SemanticSimilarityThresholdMatching",
    name="semantic similarity threshold-based matching process",
)
structural_matching = Reference(
    prefix="semapv", identifier="StructuralMatching", name="structural matching process"
)
unspecified_matching_process = Reference(
    prefix="semapv", identifier="UnspecifiedMatching", name="unspecified matching process"
)

#: A set of matching types from ``semapv``
matching_processes: set[Reference] = {
    background_knowledge_based_matching_process,
    composite_matching_process,
    instance_based_matching_process,
    lexical_matching_process,
    lexical_similarity_threshold_based_matching_process,
    logical_reasoning_matching_process,
    manual_mapping_curation,
    mapping_chaining,
    mapping_inversion,
    semantic_similarity,
    structural_matching,
    unspecified_matching_process,
}

#: See https://mapping-commons.github.io/sssom/spec-model/
match_typedefs: Sequence[Reference] = (
    broad_match,
    close_match,
    exact_match,
    narrow_match,
    related_match,
    same_as,  # for instances
    equivalent_class,  # for classes
    equivalent_property,  # for properties
    has_dbxref,
    see_also,
)

# Extension past the SSSOM spec
extended_match_typedefs: Sequence[Reference] = (
    *match_typedefs,
    alternative_term,
    term_replaced_by,
)

# Individuals

#: the author of this package. It's useful to have this reference pre-defined
#: for testing purposes
charlie = Reference(prefix="orcid", identifier="0000-0003-4423-4370", name="Charles Tapley Hoyt")



"""Trivial version test."""

import json
import tempfile
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import rdflib

import curies
from curies.api import (
    CompressionError,
    Converter,
    CURIEStandardizationError,
    DuplicatePrefixes,
    DuplicateURIPrefixes,
    ExpansionError,
    NamableReference,
    NamedReference,
    NoCURIEDelimiterError,
    PrefixStandardizationError,
    Record,
    Records,
    Reference,
    ReferenceTuple,
    URIStandardizationError,
    chain,
    upgrade_prefix_map,
)
from curies.sources import (
    BIOREGISTRY_CONTEXTS,
    get_bioregistry_converter,
    get_go_converter,
    get_monarch_converter,
    get_obo_converter,
)
from curies.version import get_version
from tests.constants import SLOW

CHEBI_URI_PREFIX = "http://purl.obolibrary.org/obo/CHEBI_"
GO_URI_PREFIX = "http://purl.obolibrary.org/obo/GO_"


class TestStruct(unittest.TestCase):
    """Test the data structures."""

    def test_not_curie(self) -> None:
        """Test a malformed CURIE."""
        with self.assertRaises(NoCURIEDelimiterError) as e:
            Reference.from_curie("not a curie")
        self.assertIn("does not appear to be a CURIE", str(e.exception))

    def test_default_prefix(self) -> None:
        """Test a default (empty) prefix."""
        ref = Reference.from_curie(":something")
        self.assertEqual("", ref.prefix)
        self.assertEqual("something", ref.identifier)

    def test_default_identifier(self) -> None:
        """Test a default (empty) identifier."""
        ref = Reference.from_curie("p1:")
        self.assertEqual("p1", ref.prefix)
        self.assertEqual("", ref.identifier)

    def test_multiple_delimiters(self) -> None:
        """Test a default (empty) identifier."""
        ref = Reference.from_curie("a1:b2:c3")
        self.assertEqual("a1", ref.prefix)
        self.assertEqual("b2:c3", ref.identifier)

    def test_records(self) -> None:
        """Test a list of records."""
        records = Records.parse_obj([{"prefix": "chebi", "uri_prefix": CHEBI_URI_PREFIX}])
        converter = Converter(records=records)
        self.assertEqual({"chebi"}, converter.get_prefixes())

    def test_sort(self) -> None:
        """Test sorting."""
        start = [
            Reference.from_curie("def:1234"),
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
        ]
        expected = [
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
            Reference.from_curie("def:1234"),
        ]
        self.assertEqual(expected, sorted(start))

    def test_set_membership(self) -> None:
        """Test membership in sets."""
        collection = {
            Reference.from_curie("def:1234"),
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
        }
        self.assertIn(Reference.from_curie("def:1234"), collection)
        self.assertNotIn(Reference.from_curie("xyz:1234"), collection)
        self.assertNotIn(Reference.from_curie(":1234"), collection)
        self.assertNotIn(Reference.from_curie("abc:"), collection)

    def test_named_set_membership(self) -> None:
        """Test membership in sets of named references."""
        references = {
            NamedReference.from_curie("a:1", "name1"),
            NamedReference.from_curie("a:2", "name2"),
        }
        self.assertIn(Reference.from_curie("a:1"), references)
        self.assertIn(NamableReference.from_curie("a:1"), references)
        self.assertIn(NamedReference.from_curie("a:1", "name1"), references)
        self.assertIn(NamableReference.from_curie("a:1", "name1"), references)
        # the following is a weird case, but shows how this works
        self.assertIn(NamedReference.from_curie("a:1", "name2"), references)

        references_2 = {
            Reference.from_curie("a:1"),
            Reference.from_curie("a:2"),
        }
        self.assertIn(Reference.from_curie("a:1"), references_2)
        self.assertIn(NamableReference.from_curie("a:1", "name1"), references_2)
        self.assertIn(NamedReference.from_curie("a:1", "name1"), references_2)


class TestAddRecord(unittest.TestCase):
    """Test adding records."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.prefix = "CHEBI"
        self.uri_prefix = CHEBI_URI_PREFIX
        self.prefix_synonym = "p"
        self.uri_prefix_synonym = "u"
        self.converter = Converter.from_extended_prefix_map(
            [
                {
                    "prefix": self.prefix,
                    "prefix_synonyms": [self.prefix_synonym],
                    "uri_prefix": self.uri_prefix,
                    "uri_prefix_synonyms": [self.uri_prefix_synonym],
                },
            ]
        )

    def test_duplicate_failure(self) -> None:
        """Test failure caused by double matching."""
        self.converter.add_prefix("GO", GO_URI_PREFIX)
        with self.assertRaises(ValueError):
            self.converter.add_record(Record(prefix="GO", uri_prefix=CHEBI_URI_PREFIX))

    def test_get_prefix_synonyms(self) -> None:
        """Test getting prefix synonyms."""
        self.assertEqual({self.prefix}, self.converter.get_prefixes())
        self.assertEqual({self.prefix}, self.converter.get_prefixes(include_synonyms=False))
        self.assertEqual(
            {self.prefix, self.prefix_synonym},
            self.converter.get_prefixes(include_synonyms=True),
        )

    def test_get_uri_prefix_synonyms(self) -> None:
        """Test getting URI prefix synonyms."""
        self.assertEqual({self.uri_prefix}, self.converter.get_uri_prefixes())
        self.assertEqual({self.uri_prefix}, self.converter.get_uri_prefixes(include_synonyms=False))
        self.assertEqual(
            {self.uri_prefix, self.uri_prefix_synonym},
            self.converter.get_uri_prefixes(include_synonyms=True),
        )

    def test_extend_on_prefix_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix="CHEBI",
                prefix_synonyms=[s1],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=["CHEBI"],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
        ]:
            with self.assertRaises(ValueError):
                self.converter.add_record(record, merge=False)
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_uri_prefix_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=s1,
                prefix_synonyms=[s3],
                uri_prefix=s2,
                uri_prefix_synonyms=[CHEBI_URI_PREFIX],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[s3],
                uri_prefix=CHEBI_URI_PREFIX,
                uri_prefix_synonyms=[s2],
            ),
        ]:
            with self.assertRaises(ValueError):
                self.converter.add_record(record, merge=False)
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, s3, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_prefix_synonym_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=self.prefix_synonym,
                prefix_synonyms=[s1],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[self.prefix_synonym],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
        ]:
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_uri_prefix_synonym_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=s1,
                prefix_synonyms=[s2],
                uri_prefix=self.uri_prefix_synonym,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[s2],
                uri_prefix=s3,
                uri_prefix_synonyms=[self.uri_prefix_synonym],
            ),
        ]:
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, s2, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_prefix_match_ci(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        record = Record(
            prefix="chebi", prefix_synonyms=[s1], uri_prefix=s2, uri_prefix_synonyms=[s3]
        )
        self.converter.add_record(record, case_sensitive=False, merge=True)
        self.assertEqual(1, len(self.converter.records))
        record = self.converter.records[0]
        self.assertEqual("CHEBI", record.prefix)
        self.assertEqual({"chebi", s1, self.prefix_synonym}, set(record.prefix_synonyms))
        self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
        self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))


class TestConverter(unittest.TestCase):
    """Test the converter class."""

    def setUp(self) -> None:
        """Set up the converter test case."""
        self.simple_obo_prefix_map = {
            "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
            "MONDO": "http://purl.obolibrary.org/obo/MONDO_",
            "GO": "http://purl.obolibrary.org/obo/GO_",
            "OBO": "http://purl.obolibrary.org/obo/",
        }
        self.converter = Converter.from_prefix_map(self.simple_obo_prefix_map)

    def test_reference_tuple(self) -> None:
        """Test the reference tuple data type."""
        t = ReferenceTuple("chebi", "1234")
        self.assertEqual("chebi:1234", t.curie)
        self.assertEqual(t, ReferenceTuple.from_curie("chebi:1234"))

    def test_reference_pydantic(self) -> None:
        """Test the reference Pydantic model."""
        t = Reference(prefix="chebi", identifier="1234")
        self.assertEqual("chebi:1234", t.curie)
        self.assertEqual(t, Reference.from_curie("chebi:1234"))

    def test_invalid_record(self) -> None:
        """Test throwing an error for invalid records."""
        with self.assertRaises(ValueError):
            Record(
                prefix="chebi",
                uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                prefix_synonyms=["chebi"],
            )
        with self.assertRaises(ValueError):
            Record(
                prefix="chebi",
                uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                uri_prefix_synonyms=["http://purl.obolibrary.org/obo/CHEBI_"],
            )

    def test_invalid_records(self) -> None:
        """Test throwing an error for duplicated URI prefixes."""
        with self.assertRaises(DuplicateURIPrefixes) as e:
            curies.load_prefix_map(
                {
                    "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                    "nope": "http://purl.obolibrary.org/obo/CHEBI_",
                }
            )
        self.assertIsInstance(str(e.exception), str)
        with self.assertRaises(DuplicatePrefixes) as e2:
            Converter(
                [
                    Record(prefix="chebi", uri_prefix="https://bioregistry.io/chebi:"),
                    Record(prefix="chebi", uri_prefix="http://purl.obolibrary.org/obo/CHEBI_"),
                ],
            )
        self.assertIsInstance(str(e2.exception), str)

        # No failure
        Converter.from_prefix_map(
            {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                "nope": "http://purl.obolibrary.org/obo/CHEBI_",
            },
            strict=False,
        )

    def test_subset(self) -> None:
        """Test subsetting a converter."""
        new_converter = self.converter.get_subconverter(["CHEBI"])
        self.assertEqual(1, len(new_converter.records))
        self.assertEqual({"CHEBI"}, new_converter.get_prefixes())
        self.assertEqual(
            {"http://purl.obolibrary.org/obo/CHEBI_"}, new_converter.get_uri_prefixes()
        )
        self.assertEqual({"CHEBI"}, set(new_converter.bimap))
        self.assertEqual({"CHEBI"}, set(new_converter.prefix_map))
        self.assertEqual(
            {"http://purl.obolibrary.org/obo/CHEBI_"}, set(new_converter.reverse_prefix_map)
        )

    def test_empty_subset(self) -> None:
        """Test subsetting a converter and getting an empty one back."""
        new_converter_2 = self.converter.get_subconverter(["NOPE"])
        self.assertEqual(0, len(new_converter_2.records))

    def test_predicates(self) -> None:
        """Add tests for predicates."""
        self.assertFalse(self.converter.is_uri(""))
        self.assertFalse(self.converter.is_uri("nope"))
        self.assertFalse(self.converter.is_curie(""))
        self.assertFalse(self.converter.is_curie("nope"))
        self.assertFalse(self.converter.is_curie(":nope"))
        self.assertFalse(self.converter.is_curie("nope:"))

    def test_convert(self) -> None:
        """Test compression."""
        self.assertEqual({"CHEBI", "MONDO", "GO", "OBO"}, self.converter.get_prefixes())
        self.assertEqual(
            {
                "http://purl.obolibrary.org/obo/CHEBI_",
                "http://purl.obolibrary.org/obo/MONDO_",
                "http://purl.obolibrary.org/obo/GO_",
                "http://purl.obolibrary.org/obo/",
            },
            self.converter.get_uri_prefixes(),
        )
        self._assert_convert(self.converter)

    def _assert_convert(self, converter: Converter) -> None:
        self.assertIn("GO", converter.prefix_map)
        self.assertIn("GO", converter.bimap)
        self.assertIn("GO", converter.reverse_bimap.values())
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.reverse_prefix_map)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.reverse_bimap)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.trie)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.bimap.values())
        for curie, uri in [
            ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
            ("OBO:unnamespaced", "http://purl.obolibrary.org/obo/unnamespaced"),
        ]:
            self.assertTrue(converter.is_uri(uri))
            self.assertTrue(converter.is_curie(curie))
            self.assertFalse(converter.is_curie(uri))
            self.assertFalse(converter.is_uri(curie))
            self.assertEqual(curie, converter.compress(uri))
            self.assertEqual(curie, converter.compress_strict(uri))
            self.assertEqual(uri, converter.expand(curie))
            self.assertEqual(uri, converter.expand_strict(curie))

        self.assertIsNone(converter.compress("http://example.org/missing:00000"))
        self.assertEqual(
            "http://example.org/missing:00000",
            converter.compress("http://example.org/missing:00000", passthrough=True),
        )
        with self.assertRaises(CompressionError):
            converter.compress_strict("http://example.org/missing:00000")

        self.assertIsNone(converter.expand("missing:00000"))
        self.assertEqual("missing:00000", converter.expand("missing:00000", passthrough=True))
        with self.assertRaises(ExpansionError):
            converter.expand_strict("missing:00000")

        self.assertLess(0, len(converter.records), msg="converter has no records")
        self.assertIsNone(converter.get_record("nope"))
        self.assertIsNone(converter.get_record("go"), msg="synonym lookup is not allowed here")
        record = converter.get_record("GO", strict=True)
        self.assertIsNotNone(record, msg=f"records: {[r.prefix for r in converter.records]}")
        self.assertIsInstance(record, Record)
        self.assertEqual("GO", record.prefix)

    @SLOW
    def test_bioregistry(self) -> None:
        """Test loading a remote JSON-LD context."""
        for web in [True, False]:
            bioregistry_converter = get_bioregistry_converter(web=web)
            self.assert_bioregistry_converter(bioregistry_converter)

        c = Converter.from_reverse_prefix_map(f"{BIOREGISTRY_CONTEXTS}/bioregistry.rpm.json")
        self.assertIn("chebi", c.prefix_map)
        self.assertNotIn("CHEBI", c.prefix_map)

    def test_jsonld(self) -> None:
        """Test parsing JSON-LD context."""
        context = {
            "@context": {
                "@version": "1.0.0",  # should skip this
                "": "",  # should skip this
                "hello": "https://example.org/hello:",
                "CHEBI": {
                    "@prefix": True,
                    "@id": "http://purl.obolibrary.org/CHEBI_",
                },
                "nope": {
                    "nope": "nope",
                },
            },
        }
        converter = Converter.from_jsonld(context)
        self.assertIn("hello", converter.prefix_map)
        self.assertIn("CHEBI", converter.prefix_map)

    @SLOW
    def test_from_github(self) -> None:
        """Test getting a JSON-LD map from GitHub."""
        with self.assertRaises(ValueError):
            # missing end .jsonld file
            Converter.from_jsonld_github("biopragmatics", "bioregistry")

        semweb_converter = Converter.from_jsonld_github(
            "biopragmatics", "bioregistry", "exports", "contexts", "semweb.context.jsonld"
        )
        self.assertIn("rdf", semweb_converter.prefix_map)

    @SLOW
    def test_obo(self) -> None:
        """Test the OBO converter."""
        obo_converter = get_obo_converter()
        self.assertIn("CHEBI", obo_converter.prefix_map)
        self.assertNotIn("chebi", obo_converter.prefix_map)

    @SLOW
    def test_monarch(self) -> None:
        """Test the Monarch converter."""
        monarch_converter = get_monarch_converter()
        self.assertIn("CHEBI", monarch_converter.prefix_map)
        self.assertNotIn("chebi", monarch_converter.prefix_map)

    @SLOW
    def test_go_registry(self) -> None:
        """Test the GO registry converter."""
        go_converter = get_go_converter()
        self.assertIn("CHEBI", go_converter.prefix_map)
        self.assertNotIn("chebi", go_converter.prefix_map)

    def assert_bioregistry_converter(self, converter: Converter) -> None:
        """Assert the bioregistry converter has the right stuff in it."""
        records = {records.prefix: records for records in converter.records}
        self.assertIn("chebi", records)
        record = records["chebi"]
        self.assertIsInstance(record, Record)
        self.assertEqual("chebi", record.prefix)
        self.assertIn("CHEBI", record.prefix_synonyms)
        self.assertIn("ChEBI", record.prefix_synonyms)

        self.assertIn("chebi", converter.prefix_map)
        self.assertIn("chebi", converter.bimap)
        # Synonyms that are non-conflicting also get added
        self.assertIn("CHEBI", converter.prefix_map)
        self.assertNotIn("CHEBI", converter.bimap)
        chebi_uri = converter.prefix_map["chebi"]
        self.assertIn(chebi_uri, converter.reverse_prefix_map)
        self.assertEqual("chebi", converter.reverse_prefix_map[chebi_uri])

    def test_load_path(self) -> None:
        """Test loading from paths."""
        with tempfile.TemporaryDirectory() as directory:
            path = Path(directory).joinpath("pm.json")
            with self.assertRaises(FileNotFoundError):
                Converter.from_prefix_map(path)
            with self.assertRaises(FileNotFoundError):
                Converter.from_prefix_map(str(path))

            path.write_text(json.dumps(self.converter.prefix_map))

            c1 = Converter.from_prefix_map(path)
            self.assertEqual(self.converter.prefix_map, c1.prefix_map)

            c2 = Converter.from_prefix_map(str(path))
            self.assertEqual(self.converter.prefix_map, c2.prefix_map)

    def test_reverse_constructor(self) -> None:
        """Test constructing from a reverse prefix map."""
        converter = Converter.from_reverse_prefix_map(
            {
                "http://purl.obolibrary.org/obo/CHEBI_": "CHEBI",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=": "CHEBI",
                "http://purl.obolibrary.org/obo/MONDO_": "MONDO",
            }
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488", converter.expand("CHEBI:138488")
        )

        self.assertEqual(
            "CHEBI:138488", converter.compress("http://purl.obolibrary.org/obo/CHEBI_138488")
        )
        self.assertEqual(
            "CHEBI:138488",
            converter.compress("https://www.ebi.ac.uk/chebi/searchId.do?chebiId=138488"),
        )

    def test_standardize_curie(self) -> None:
        """Test standardize CURIE."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=[
                        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:",
                    ],
                ),
            ]
        )
        self.assertEqual("CHEBI:138488", converter.standardize_curie("chebi:138488"))
        self.assertEqual("CHEBI:138488", converter.standardize_curie("CHEBI:138488"))
        self.assertIsNone(converter.standardize_curie("NOPE:NOPE"))
        self.assertEqual("NOPE:NOPE", converter.standardize_curie("NOPE:NOPE", passthrough=True))
        with self.assertRaises(CURIEStandardizationError):
            converter.standardize_curie("NOPE:NOPE", strict=True)

        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.standardize_uri(
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:138488"
            ),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.standardize_uri("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertIsNone(converter.standardize_uri("NOPE"))
        self.assertEqual("NOPE", converter.standardize_uri("NOPE", passthrough=True))
        with self.assertRaises(URIStandardizationError):
            converter.standardize_uri("NOPE:NOPE", strict=True)

    def test_combine(self) -> None:
        """Test chaining converters."""
        with self.assertRaises(ValueError):
            chain([])

        c1 = Converter.from_priority_prefix_map(
            {
                "CHEBI": ["http://purl.obolibrary.org/obo/CHEBI_", "https://bioregistry.io/chebi:"],
                "MONDO": ["http://purl.obolibrary.org/obo/MONDO_"],
            }
        )
        c2 = Converter.from_priority_prefix_map(
            {
                "CHEBI": [
                    "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=",
                    "http://identifiers.org/chebi/",
                    "http://purl.obolibrary.org/obo/CHEBI_",
                ],
                "GO": ["http://purl.obolibrary.org/obo/GO_"],
                "OBO": ["http://purl.obolibrary.org/obo/"],
            }
        )
        converter = chain([c1, c2], case_sensitive=True)

        self.assertEqual("CHEBI", converter.get_record("CHEBI").prefix)
        for url in [
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            "https://bioregistry.io/chebi:138488",
            "http://identifiers.org/chebi/138488",
            "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=138488",
        ]:
            self.assertEqual("CHEBI:138488", converter.compress(url))

        self.assertEqual("GO", converter.get_record("GO").prefix)
        self.assertEqual(
            "GO:0000001",
            converter.compress("http://purl.obolibrary.org/obo/GO_0000001"),
        )

        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_", converter.get_record("CHEBI").uri_prefix
        )
        self.assertIn("CHEBI", converter.prefix_map)
        self.assertEqual("http://purl.obolibrary.org/obo/CHEBI_", converter.prefix_map["CHEBI"])
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand("CHEBI:138488"),
        )
        self.assertNotIn("nope", converter.get_prefixes())

    def test_combine_with_synonyms(self) -> None:
        """Test combination with synonyms."""
        r1 = Record(prefix="GO", uri_prefix=GO_URI_PREFIX)
        r2 = Record(prefix="go", prefix_synonyms=["GO"], uri_prefix="https://identifiers.org/go:")

        c1 = Converter([])
        c1.add_record(r1)
        self.assertEqual(c1.records, Converter([r1]).records)

        c1.add_record(r2, merge=True)
        self.assertEqual(1, len(c1.records))
        r = c1.records[0]
        self.assertEqual("GO", r.prefix)
        self.assertEqual({"go"}, set(r.prefix_synonyms))
        self.assertEqual("http://purl.obolibrary.org/obo/GO_", r.uri_prefix)
        self.assertEqual({"https://identifiers.org/go:"}, set(r.uri_prefix_synonyms))

        c3 = chain([Converter([r1]), Converter([r2])])
        self.assertEqual(1, len(c3.records))
        self.assertIn("GO", c3.prefix_map)
        self.assertIn("go", c3.prefix_map, msg=f"PM: {c3.prefix_map}")
        self.assertNotIn("go", c3.bimap)
        self.assertIn("GO", c3.bimap)

    def test_combine_ci(self) -> None:
        """Test combining case-insensitive."""
        c1 = Converter.from_priority_prefix_map(
            {
                "CHEBI": [
                    "http://purl.obolibrary.org/obo/CHEBI_",
                    "https://bioregistry.io/chebi:",
                ],
            }
        )
        c2 = Converter.from_reverse_prefix_map(
            {
                "http://identifiers.org/chebi/": "chebi",
                "http://identifiers.org/chebi:": "chebi",
            }
        )
        converter = chain([c1, c2], case_sensitive=False)
        self.assertEqual({"CHEBI"}, converter.get_prefixes())
        self.assertEqual({"CHEBI"}, converter.get_prefixes(include_synonyms=False))
        self.assertEqual({"CHEBI", "chebi"}, converter.get_prefixes(include_synonyms=True))
        for url in [
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            "http://identifiers.org/chebi/138488",
            "http://identifiers.org/chebi:138488",
            "https://bioregistry.io/chebi:138488",
        ]:
            self.assertEqual("CHEBI:138488", converter.compress(url))
        # use the first prefix map for expansions
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand("CHEBI:138488"),
        )

    def test_combine_with_patterns(self) -> None:
        """Test chaining with patterns."""
        c1 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")])
        c2 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d+")])
        converter = chain([c1, c2])
        self.assertEqual(
            [Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")],
            converter.records,
        )

    def test_combine_with_patterns_via_synonym(self) -> None:
        """Test chaining with patterns."""
        c1 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")])
        c2 = Converter(
            [
                Record(
                    prefix="b",
                    prefix_synonyms=["a"],
                    uri_prefix="https://example.org/b/",
                    pattern="^\\d+",
                )
            ]
        )
        converter = chain([c1, c2])
        self.assertEqual(
            [
                Record(
                    prefix="a",
                    prefix_synonyms=["b"],
                    uri_prefix="https://example.org/a/",
                    uri_prefix_synonyms=["https://example.org/b/"],
                    pattern="^\\d{7}",
                )
            ],
            converter.records,
        )

    def test_df_bulk(self) -> None:
        """Test bulk processing in pandas dataframes."""
        rows = [
            ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
        ]
        df = pd.DataFrame(rows, columns=["curie", "uri"])
        self.converter.pd_expand(df, "curie")
        self.assertTrue((df.curie == df.uri).all())

        df = pd.DataFrame(rows, columns=["curie", "uri"])
        self.converter.pd_compress(df, "uri")
        self.assertTrue((df.curie == df.uri).all())

    def test_df_standardize(self) -> None:
        """Test standardizing dataframes."""
        converter = Converter([])
        converter.add_prefix(
            "chebi",
            "http://purl.obolibrary.org/obo/CHEBI_",
            prefix_synonyms=["CHEBI"],
            uri_prefix_synonyms=["https://bioregistry.io/chebi:"],
        )
        self.assertEqual("chebi", converter.standardize_prefix("chebi"))
        self.assertEqual("chebi", converter.standardize_prefix("CHEBI"))
        self.assertIsNone(converter.standardize_prefix("nope"))
        self.assertEqual("nope", converter.standardize_prefix("nope", passthrough=True))
        with self.assertRaises(PrefixStandardizationError):
            converter.standardize_prefix("nope", strict=True)

        rows = [
            ("chebi", "CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
            ("CHEBI", "CHEBI:2", "https://bioregistry.io/chebi:2"),
        ]
        df = pd.DataFrame(rows, columns=["prefix", "curie", "uri"])
        converter.pd_standardize_prefix(df, column="prefix")
        self.assertEqual(["chebi", "chebi"], list(df["prefix"]), msg=f"\n\n{df}")

        converter.pd_standardize_curie(df, column="curie")
        self.assertEqual(["chebi:1", "chebi:2"], list(df["curie"]))

        converter.pd_standardize_uri(df, column="uri")
        self.assertEqual(
            ["http://purl.obolibrary.org/obo/CHEBI_1", "http://purl.obolibrary.org/obo/CHEBI_2"],
            list(df["uri"]),
        )

    def test_file_bulk(self) -> None:
        """Test bulk processing of files."""
        with TemporaryDirectory() as directory:
            for rows, header in [
                (
                    [
                        ("curie", "uri"),
                        ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
                    ],
                    True,
                ),
                (
                    [
                        ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
                    ],
                    False,
                ),
            ]:
                path = Path(directory).joinpath("test.tsv")
                with path.open("w") as file:
                    for row in rows:
                        print(*row, sep="\t", file=file)

                idx = 1 if header else 0

                self.converter.file_expand(path, 0, header=header)
                lines = [line.strip().split("\t") for line in path.read_text().splitlines()]
                self.assertEqual("http://purl.obolibrary.org/obo/CHEBI_1", lines[idx][0])

                self.converter.file_compress(path, 0, header=header)
                lines = [line.strip().split("\t") for line in path.read_text().splitlines()]
                self.assertEqual("CHEBI:1", lines[idx][0])

    def test_incremental(self) -> None:
        """Test building a converter from an incremental interface."""
        converter = Converter([])
        for prefix, uri_prefix in self.simple_obo_prefix_map.items():
            converter.add_prefix(prefix, uri_prefix)
        converter.add_prefix(
            "hgnc",
            "https://bioregistry.io/hgnc:",
            prefix_synonyms=["HGNC"],
            uri_prefix_synonyms=["https://identifiers.org/hgnc:"],
        )
        self._assert_convert(converter)
        self.assertEqual(
            "hgnc:1234",
            converter.compress("https://bioregistry.io/hgnc:1234"),
        )
        self.assertEqual(
            "hgnc:1234",
            converter.compress("https://identifiers.org/hgnc:1234"),
        )
        self.assertEqual("https://bioregistry.io/hgnc:1234", converter.expand("HGNC:1234"))

        with self.assertRaises(ValueError):
            converter.add_prefix("GO", "...")
        with self.assertRaises(ValueError):
            converter.add_prefix("...", "http://purl.obolibrary.org/obo/GO_")
        with self.assertRaises(ValueError):
            converter.add_prefix(
                "...", "...", uri_prefix_synonyms=["http://purl.obolibrary.org/obo/GO_"]
            )
        with self.assertRaises(ValueError):
            converter.add_prefix("...", "...", prefix_synonyms=["GO"])

    def test_rdflib(self) -> None:
        """Test parsing a converter from an RDFLib object."""
        graph = rdflib.Graph()
        for prefix, uri_prefix in self.simple_obo_prefix_map.items():
            graph.bind(prefix, uri_prefix)
        converter = Converter.from_rdflib(graph)
        self._assert_convert(converter)

        converter_2 = Converter.from_rdflib(graph.namespace_manager)
        self._assert_convert(converter_2)

    def test_expand_all(self) -> None:
        """Test expand all."""
        priority_prefix_map = {
            "CHEBI": [
                "http://purl.obolibrary.org/obo/CHEBI_",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:",
            ],
        }
        converter = Converter.from_priority_prefix_map(priority_prefix_map)
        self.assertEqual(
            [
                "http://purl.obolibrary.org/obo/CHEBI_138488",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:138488",
            ],
            converter.expand_all("CHEBI:138488"),
        )
        self.assertIsNone(converter.expand_all("NOPE:NOPE"))

    def test_expand_ambiguous(self) -> None:
        """Test expansion of URI or CURIEs."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=["https://identifiers.org/chebi:"],
                ),
            ]
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("CHEBI:138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("chebi:138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("https://identifiers.org/chebi:138488"),
        )

        self.assertIsNone(converter.expand_or_standardize("missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("missing:0000000", strict=True)
        self.assertEqual(
            "missing:0000000", converter.expand_or_standardize("missing:0000000", passthrough=True)
        )

        self.assertIsNone(converter.expand_or_standardize("https://example.com/missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("https://example.com/missing:0000000", strict=True)
        self.assertEqual(
            "https://example.com/missing:0000000",
            converter.expand_or_standardize(
                "https://example.com/missing:0000000", passthrough=True
            ),
        )

    def test_compress_ambiguous(self) -> None:
        """Test compression of URI or CURIEs."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=["https://identifiers.org/chebi:"],
                ),
            ]
        )
        self.assertEqual("CHEBI:138488", converter.compress_or_standardize("CHEBI:138488"))
        self.assertEqual("CHEBI:138488", converter.compress_or_standardize("chebi:138488"))
        self.assertEqual(
            "CHEBI:138488",
            converter.compress_or_standardize("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertEqual(
            "CHEBI:138488",
            converter.compress_or_standardize("https://identifiers.org/chebi:138488"),
        )

        self.assertIsNone(converter.expand_or_standardize("missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("missing:0000000", strict=True)
        self.assertEqual(
            "missing:0000000", converter.expand_or_standardize("missing:0000000", passthrough=True)
        )

        self.assertIsNone(converter.expand_or_standardize("https://example.com/missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("https://example.com/missing:0000000", strict=True)
        self.assertEqual(
            "https://example.com/missing:0000000",
            converter.expand_or_standardize(
                "https://example.com/missing:0000000", passthrough=True
            ),
        )


class TestVersion(unittest.TestCase):
    """Trivially test a version."""

    def test_version_type(self) -> None:
        """Test the version is a string.

        This is only meant to be an example test.
        """
        version = get_version()
        self.assertIsInstance(version, str)


class TestUtils(unittest.TestCase):
    """Test utility functions."""

    def test_clean(self) -> None:
        """Test clean."""
        prefix_map = {
            "b": "https://example.com/a/",
            "a": "https://example.com/a/",
            "c": "https://example.com/c/",
        }
        records = upgrade_prefix_map(prefix_map)
        self.assertEqual(2, len(records))
        a_record, c_record = records

        self.assertEqual("a", a_record.prefix)
        self.assertEqual(["b"], a_record.prefix_synonyms)
        self.assertEqual("https://example.com/a/", a_record.uri_prefix)
        self.assertEqual([], a_record.uri_prefix_synonyms)

        self.assertEqual("c", c_record.prefix)
        self.assertEqual([], c_record.prefix_synonyms)
        self.assertEqual("https://example.com/c/", c_record.uri_prefix)
        self.assertEqual([], c_record.uri_prefix_synonyms)



"""Test W3C validation."""

import unittest
from pathlib import Path

from curies.w3c import is_w3c_curie, is_w3c_prefix

HERE = Path(__file__).parent.resolve()
DIRECTORY = HERE.joinpath("resources")
VALID_CURIES_PATH = DIRECTORY.joinpath("valid_curies.txt")
INVALID_CURIES_PATH = DIRECTORY.joinpath("invalid_curies.txt")
VALID_PREFIXES_PATH = DIRECTORY.joinpath("valid_prefixes.txt")
INVALID_PREFIXES_PATH = DIRECTORY.joinpath("invalid_prefixes.txt")


def _read(path: Path) -> list[str]:
    return path.read_text().splitlines()


class TestValidators(unittest.TestCase):
    """Test W3C validation."""

    def test_prefixes(self) -> None:
        """Test prefixes validation."""
        for prefix in _read(VALID_PREFIXES_PATH):
            with self.subTest(prefix=prefix):
                self.assertTrue(is_w3c_prefix(prefix))

        for prefix in _read(INVALID_PREFIXES_PATH):
            with self.subTest(prefix=prefix):
                self.assertFalse(is_w3c_prefix(prefix))

    def test_curies(self) -> None:
        """Test CURIE validation."""
        for curie in _read(VALID_CURIES_PATH):
            with self.subTest(curie=curie):
                self.assertTrue(is_w3c_curie(curie), msg="CURIE should test as valid, but did not")

        for curie in _read(INVALID_CURIES_PATH):
            with self.subTest(curie=curie):
                self.assertFalse(
                    is_w3c_curie(curie), msg="CURIE should test as invalid, but did not"
                )



"""Test discovering a prefix map from a list of URIs."""

import unittest
from typing import ClassVar

import rdflib

from curies import Converter, Record
from curies.discovery import discover, discover_from_rdf
from tests.constants import SLOW


class TestDiscovery(unittest.TestCase):
    """Test discovery of URI prefixes."""

    converter: ClassVar[Converter]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case with a dummy converter."""
        cls.converter = Converter(
            [
                Record(prefix="GO", uri_prefix="http://purl.obolibrary.org/obo/GO_"),
                Record(prefix="rdfs", uri_prefix=str(rdflib.RDFS._NS)),
            ]
        )

    def test_simple(self) -> None:
        """Test a simple case of discovering URI prefixes."""
        uris = [f"http://ran.dom/{i:03}" for i in range(30)]
        uris.append("http://purl.obolibrary.org/obo/GO_0001234")

        converter = discover(uris, cutoff=3, converter=self.converter)
        self.assertEqual([Record(prefix="ns1", uri_prefix="http://ran.dom/")], converter.records)
        self.assertEqual("ns1:001", converter.compress("http://ran.dom/001"))
        self.assertIsNone(
            converter.compress("http://purl.obolibrary.org/obo/GO_0001234"),
            msg="discovered converter should not inherit reference converter's definitions",
        )

        converter = discover(uris, cutoff=50, converter=self.converter)
        self.assertEqual([], converter.records)
        self.assertIsNone(
            converter.compress("http://ran.dom/001"),
            msg="cutoff was high, so discovered converter should not detect `http://ran.dom/`",
        )

    def test_rdflib(self) -> None:
        """Test discovery in RDFlib."""
        graph = rdflib.Graph()
        for i in range(30):
            graph.add(
                (
                    rdflib.URIRef(f"http://ran.dom/{i:03}"),
                    rdflib.RDFS.subClassOf,
                    rdflib.URIRef(f"http://ran.dom/{i + 1:03}"),
                )
            )
            graph.add(
                (
                    rdflib.URIRef(f"http://ran.dom/{i:03}"),
                    rdflib.RDFS.label,
                    rdflib.Literal(f"Node {i}"),
                )
            )

        converter = discover_from_rdf(graph, converter=self.converter)
        self.assertEqual([Record(prefix="ns1", uri_prefix="http://ran.dom/")], converter.records)
        self.assertEqual("ns1:001", converter.compress("http://ran.dom/001"))
        self.assertIsNone(
            converter.compress("http://purl.obolibrary.org/obo/GO_0001234"),
            msg="discovered converter should not inherit reference converter's definitions",
        )

    @SLOW
    def test_remote(self) -> None:
        """Test parsing AEON."""
        converter = discover_from_rdf(
            graph="https://raw.githubusercontent.com/tibonto/aeon/main/aeon.owl",
            format="xml",
        )
        self.assertIn("http://purl.obolibrary.org/obo/AEON_", converter.reverse_prefix_map)



"""Utilities for the mapping service."""

from __future__ import annotations

import json
import json.decoder
import unittest
from collections.abc import Mapping
from typing import Callable

from defusedxml import ElementTree

__all__ = [
    "CONTENT_TYPE_TO_HANDLER",
    "get_sparql_records",
    "handle_csv",
    "handle_header",
    "handle_json",
    "handle_xml",
    "parse_header",
    "sparql_service_available",
]

Record = Mapping[str, str]
Records = list[Record]

#: A SPARQL query used to ping a SPARQL endpoint
PING_SPARQL = 'SELECT ?s ?o WHERE { BIND("hello" as ?s) . BIND("there" as ?o) . }'

#: This is default for federated queries
DEFAULT_CONTENT_TYPE = "application/sparql-results+xml"

#: A mapping from content types to the keys used for serializing
#: in :meth:`rdflib.Graph.serialize` and other serialization functions
CONTENT_TYPE_TO_RDFLIB_FORMAT = {
    # https://www.w3.org/TR/sparql11-results-json/
    "application/sparql-results+json": "json",
    # https://www.w3.org/TR/rdf-sparql-XMLres/
    "application/sparql-results+xml": "xml",
    # https://www.w3.org/TR/sparql11-results-csv-tsv/
    "application/sparql-results+csv": "csv",
}

#: A dictionary that maps synonym content types to the canonical ones
CONTENT_TYPE_SYNONYMS = {
    "application/json": "application/sparql-results+json",
    "text/json": "application/sparql-results+json",
    "application/xml": "application/sparql-results+xml",
    "text/xml": "application/sparql-results+xml",
    "text/csv": "application/sparql-results+csv",
}


def handle_json(text: str) -> Records:
    """Parse bindings encoded in a JSON string."""
    data = json.loads(text)
    return [
        {key: value["value"] for key, value in record.items()}
        for record in data["results"]["bindings"]
    ]


def handle_xml(text: str) -> Records:
    """Parse bindings encoded in an XML string."""
    root = ElementTree.fromstring(text)
    results = root.find("{http://www.w3.org/2005/sparql-results#}results")
    return [
        {
            binding.attrib["name"]: binding.find("{http://www.w3.org/2005/sparql-results#}uri").text
            for binding in result
        }
        for result in results
    ]


def handle_csv(text: str) -> Records:
    """Parse bindings encoded in a CSV string."""
    header, *lines = (line.strip().split(",") for line in text.splitlines())
    return [dict(zip(header, line)) for line in lines]


#: A mapping from canonical content types to functions for parsing them
CONTENT_TYPE_TO_HANDLER: Mapping[str, Callable[[str], Records]] = {
    "application/sparql-results+json": handle_json,
    "application/sparql-results+xml": handle_xml,
    "application/sparql-results+csv": handle_csv,
}


def get_sparql_records(endpoint: str, sparql: str, accept: str) -> Records:
    """Get a response from a given SPARQL query."""
    import requests

    res = requests.get(
        endpoint,
        timeout=60,
        params={"query": sparql},
        headers={"accept": accept},
    )
    res.raise_for_status()
    func = CONTENT_TYPE_TO_HANDLER[handle_header(accept)]
    return func(res.text)


def get_sparql_record_so_tuples(records: Records) -> set[tuple[str, str]]:
    """Get subject/object pairs from records."""
    return {(record["s"], record["o"]) for record in records}


def sparql_service_available(endpoint: str) -> bool:
    """Test if a SPARQL service is running."""
    try:
        records = get_sparql_records(endpoint, PING_SPARQL, "application/json")
    except (OSError, json.decoder.JSONDecodeError):
        return False
    return {("hello", "there")} == get_sparql_record_so_tuples(records)


def _handle_part(part: str) -> tuple[str, float]:
    if ";q=" not in part:
        return part, 1.0
    key, q = part.split(";q=", 1)
    return key, float(q)


def parse_header(header: str) -> list[str]:
    """Parse the header and sort in descending order of q value."""
    parts = dict(_handle_part(part) for part in header.split(","))
    return sorted(parts, key=parts.__getitem__, reverse=True)


def handle_header(header: str | None, default: str = DEFAULT_CONTENT_TYPE) -> str:
    """Canonicalize a header."""
    if not header:
        return default

    for header_part in parse_header(header):
        header_part = CONTENT_TYPE_SYNONYMS.get(header_part, header_part)
        if header_part in CONTENT_TYPE_TO_RDFLIB_FORMAT:
            return header_part
        # What happens if encountering "*/*" that has a higher q than something else?
        # Is that even possible/coherent?

    return default


def require_service(url: str, name: str):  # type:ignore
    """Skip a test unless the service is available."""
    return unittest.skipUnless(
        sparql_service_available(url), reason=f"No {name} service is running on {url}"
    )



"""Tests federated SPARQL queries with a locally deployed triple store."""

import itertools as itt
import time
import unittest
from multiprocessing import Process
from typing import ClassVar

import fastapi
import uvicorn

from curies import Converter
from curies.mapping_service import get_fastapi_mapping_app
from curies.mapping_service.utils import (
    get_sparql_record_so_tuples,
    get_sparql_records,
    require_service,
    sparql_service_available,
)
from tests.test_mapping_service import PREFIX_MAP

BLAZEGRAPH_ENDPOINT = "http://localhost:9999/blazegraph/namespace/kb/sparql"
BLAZEGRAPH_JAR_URL = (
    "https://github.com/blazegraph/database/releases/download/BLAZEGRAPH_2_1_6_RC/blazegraph.jar"
)
SPARQL_VALUES_FMT = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    SERVICE <{mapping_service}> {{
        VALUES ?s {{ <http://purl.obolibrary.org/obo/CHEBI_24867> }}
        ?s owl:sameAs ?o
    }}
}}
""".rstrip()

SPARQL_VALUES_FMT_2 = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    VALUES ?s {{ <http://purl.obolibrary.org/obo/CHEBI_24867> }}
    SERVICE <{mapping_service}> {{
        ?s owl:sameAs ?o
    }}
}}
""".rstrip()

SPARQL_VALUES_FMT_3 = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    SERVICE <{mapping_service}> {{
        ?s owl:sameAs ?o
    }}
    VALUES ?s {{ <http://purl.obolibrary.org/obo/CHEBI_24867> }}
}}
""".rstrip()

SPARQL_SIMPLE_FMT = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    SERVICE <{mapping_service}> {{
        <http://purl.obolibrary.org/obo/CHEBI_24867> owl:sameAs ?o .
        ?s owl:sameAs ?o .
    }}
}}
""".rstrip()

SPARQL_SIMPLE_FMT_2 = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    SERVICE <{mapping_service}> {{
        ?s owl:sameAs ?o .
        <http://purl.obolibrary.org/obo/CHEBI_24867> owl:sameAs ?o .
    }}
}}
""".rstrip()

# TODO test
SPARQL_SIMPLE_FMT_3 = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    <http://purl.obolibrary.org/obo/CHEBI_24867> owl:sameAs ?o .
    SERVICE <{mapping_service}> {{
        ?s owl:sameAs ?o .
    }}
}}
""".rstrip()

# TODO test
SPARQL_SIMPLE_FMT_4 = """\
PREFIX owl: <http://www.w3.org/2002/07/owl#>
SELECT DISTINCT ?s ?o WHERE {{
    SERVICE <{mapping_service}> {{
        ?s owl:sameAs ?o .
    }}
    <http://purl.obolibrary.org/obo/CHEBI_24867> owl:sameAs ?o .
}}
""".rstrip()


class FederationMixin(unittest.TestCase):
    """A shared mixin for testing."""

    def assert_service_works(self, endpoint: str) -> None:
        """Assert that a service is able to accept a simple SPARQL query."""
        self.assertTrue(sparql_service_available(endpoint))


def _get_app() -> fastapi.FastAPI:
    converter = Converter.from_priority_prefix_map(PREFIX_MAP)
    app = get_fastapi_mapping_app(converter)
    return app


@require_service(BLAZEGRAPH_ENDPOINT, "Blazegraph")
class TestFederatedSparql(FederationMixin):
    """Test the identifier mapping service."""

    endpoint: ClassVar[str] = BLAZEGRAPH_ENDPOINT
    mimetypes: ClassVar[list[str]] = [
        "application/sparql-results+json",
        "application/sparql-results+xml",
        "text/csv",  # for some reason, Blazegraph wants this instead of application/sparql-results+csv
    ]
    query_formats: ClassVar[list[str]] = [
        SPARQL_VALUES_FMT,
        SPARQL_VALUES_FMT_2,
        SPARQL_VALUES_FMT_3,
        SPARQL_SIMPLE_FMT,
        SPARQL_SIMPLE_FMT_2,
    ]
    host: ClassVar[str] = "localhost"
    port: ClassVar[int] = 8000
    mapping_service_process: Process

    def setUp(self) -> None:
        """Set up the test case."""
        # Start the curies mapping service SPARQL endpoint
        self.mapping_service_process = Process(
            target=uvicorn.run,
            # uvicorn.run accepts a zero-argument callable that returns an app
            args=(_get_app,),
            kwargs={"host": self.host, "port": self.port, "log_level": "info"},
            daemon=True,
        )
        self.mapping_service_process.start()
        time.sleep(5)

        self.mapping_service = f"http://{self.host}:{self.port}/sparql"
        self.queries = [
            query_format.format(mapping_service=self.mapping_service)
            for query_format in self.query_formats
        ]

        self.assert_service_works(self.mapping_service)

    def tearDown(self) -> None:
        """Tear down the testing case."""
        self.mapping_service_process.kill()

    def test_federated_local(self) -> None:
        """Test sending a federated query to a local mapping service from a local service."""
        for mimetype, sparql in itt.product(self.mimetypes, self.queries):
            with self.subTest(mimetype=mimetype, sparql=sparql):
                records = get_sparql_records(self.endpoint, sparql, accept=mimetype)
                self.assertIn(
                    (
                        "http://purl.obolibrary.org/obo/CHEBI_24867",
                        "http://identifiers.org/chebi/24867",
                    ),
                    get_sparql_record_so_tuples(records),
                )
